#!/usr/bin/python
# -*- coding: UTF-8 -*-
from sklearn.datasets import load_wine
from sklearn.model_selection import train_test_split
import numpy as np

feature_name = ['酒精', '苹果酸', '灰', '灰的碱性', '镁', '总酚', '类黄酮',
                '非黄烷类酚类', '花青素', '颜色强度', '色调', 'od280/od315稀释葡萄酒', '脯氨酸']
class_names=["琴酒", "雪莉", "贝尔摩德"]

# 生成决策树的节点类型
class TreeNode(object):
    idx = 0                  # 属性
    idx_value = 0.0          # 属性值
    is_leaf = False
    y = 0                   # 预测值
    samples = 0             # 样本数
    value = []              # 分布情况
    left = None
    right = None

    def __init__(self, idx,idx_value, is_leaf, y, samples, value, left=None, right=None):
        self.idx = idx
        self.idx_value = idx_value
        self.is_leaf = is_leaf
        self.y = y
        self.samples = samples
        self.value = value
        self.left = left
        self.right = right

    def __str__(self):
        return "%s,%f, leaf:%d, samples:%d, value:%s, %s" % \
               (feature_name[self.idx], self.idx_value, self.is_leaf,self.samples, self.value, class_names[self.y])


def read_data():
    wine = load_wine()
    print(wine.data.shape)  # 178*13
    print(wine.target)
    print(wine.feature_names)
    print(wine.target_names)
    # 如果wine是一张表，应该长这样：
    import pandas as pd
    # pdata = pd.concat([pd.DataFrame(wine.data), pd.DataFrame(wine.target)], axis=1)
    Xtrain, Xtest, Ytrain, Ytest = train_test_split(wine.data,wine.target,test_size=0.3)
    return Xtrain, Xtest, Ytrain, Ytest


def calc_ent(x):
    """
        calculate shanno ent of x
    """
    x_value_list = set([x[i] for i in range(x.shape[0])])
    ent = 0.0
    for x_value in x_value_list:
        p = float(x[x == x_value].shape[0]) / x.shape[0]
        logp = np.log2(p)
        ent -= p * logp

    return ent


def calc_ent1(x):
    """
        calculate shanno ent of x
    """
    p_set = []
    for item in x:
        for i in p_set:
            if i[0] == item:
                i[1] = i[1] + 1
                break
        else:
            i = [item, 1]
            p_set.append(i)

    pro_list = []
    size = len(x)
    for item in p_set:
        pro_list.append(item[1]/size)

    ent = 0.0
    for p in pro_list:
        logp = np.log2(p)
        ent -= p * logp

    return ent


# 计算某个属性的信息增益
def cal_ent_attr(Xtrain, Ytrain):
    print('sharp', Xtrain.shape)

    # 对每个属性
    min_ent = 100
    min_i = 0
    min_mean = 0
    for i in range(Xtrain.shape[1]):
        x_value_list = set([Xtrain[j][i] for j in range(Xtrain.shape[0])])
        mean = sum(x_value_list)/len(x_value_list)
        sum_ent = 0
        # 二叉树
        p = Ytrain[Xtrain[:, i] > mean]
        sum_ent = sum_ent + calc_ent(p)
        p = Ytrain[Xtrain[:, i] <= mean]
        sum_ent = sum_ent + calc_ent(p)

        if sum_ent < min_ent:
            min_ent = sum_ent
            min_i = i
            min_mean = mean
    return min_i, min_mean


MAX_T = 5


def is_end(Ytrain):
    if len(Ytrain) == 0:
        return True
    if len(set(Ytrain)) == 1:
        return True

# 强行划分为叶子节点
def leaf_node(Ytrain):
    p_set = []
    for item in Ytrain:
        for i in p_set:
            if i[0] == item:
                i[1] = i[1] + 1
                break
        else:
            i = [item, 1]
            p_set.append(i)

    max_item = [0, 0]
    for item in p_set:
        if item[1] > max_item[1]:
            max_item = item
    print('这个是叶子节点，value:', max_item[0])
    return TreeNode(0, 0, True, max_item[0], 0, 0)


# def create_node()


def fit(Xtrain, Ytrain, parent_node, depth):

    if is_end(Ytrain):
        print('这个是叶子节点')
        return TreeNode(0, 0, True, 0, 0, 0)

    if depth > MAX_T:
        return leaf_node(Ytrain)

    i, mean = cal_ent_attr(Xtrain, Ytrain)
    print("第", i, "个属性,mean:", mean)
    # 生成节点
    parent_node = TreeNode(i, mean, False, 0, 0, 0)

    # 切分数据
    right_Ytrain = Ytrain[Xtrain[:, i] > mean]
    right_Xtrain = Xtrain[Xtrain[:, i] > mean]
    # right_Xtrain = np.delete(right_Xtrain, i, axis=1) # 这个属性还可以再被切分

    right_node = fit(right_Xtrain, right_Ytrain, parent_node, depth+1)

    left_Ytrain = Ytrain[Xtrain[:, i] <= mean]
    left_Xtrain = Xtrain[Xtrain[:, i] <= mean]
    # left_Xtrain = np.delete(left_Xtrain, i, axis=1)
    left_node = fit(left_Xtrain, left_Ytrain, parent_node, depth + 1)

    parent_node.left = left_node
    parent_node.right = right_node
    return parent_node


def print_width(nodes, depth):
    if len(nodes) == 0:
        return

    print("--第", depth, "层--")
    node_down = []
    for node in nodes:
        print(node)
        if node.left is not None:
            node_down.append(node.left)
        if node.right is not None:
            node_down.append(node.right)


    print_width(node_down, depth+1)


if __name__ == '__main__':
    Xtrain, Xtest, Ytrain, Ytest = read_data()
    print(calc_ent1(Ytrain))
    print(calc_ent(Ytrain))

    print(cal_ent_attr(Xtrain, Ytrain))

    node = fit(Xtrain, Ytrain, None, 0)
    print_width([node], 1)
